<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Version details
 *
 * @package    block_iclicker
 * @copyright  2012 i>clicker (R) <http://www.iclicker.com/dnn/> onwards Saji Nair (http://sajinair.com)
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

/**
 * Handles controller functions related to the views
 */

require_once('../../config.php');
global $CFG, $USER, $COURSE;
require_once('iclicker_service.php');

class iclicker_controller {

    // constants
    const TYPE_HTML = 'html';
    const TYPE_XML = 'xml';
    const TYPE_TEXT = 'txt';

    const PASSWORD = '_password';
    const LOGIN = '_login';
    const SSO_KEY = '_key';
    const SEPARATOR = '/';
    const PERIOD = '.';
    const XML_HEADER = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n";
    const SESSION_ID = '_sessionId';
    const COMPENSATE_METHOD = '_method';
    const XML_DATA = '_xml';

    // class vars

    // REQUEST
    public $method = 'GET';
    public $path = '';
    public $query = '';
    public $body = null;

    // RESPONSE
    public $status = 200;
    public $message = '';
    public $headers = array();

    public $results = array(
    );

    public function __construct($getbody = false) {
        // set some headers
        $this->headers['Content-Encoding'] = 'UTF8';
        // header('Content-type: text/plain');
        // header('Cache-Control: no-cache, must-revalidate');
        // get the rest path
        $fullpath = me();
        if ($fullpath) {
            $path = $fullpath;
            $pos = strripos($fullpath, '.php');
            if ($pos > 1) {
                $path = substr($fullpath, $pos + 4);
                $path = trim($path, '/ '); // trim whitespace and slashes
            }
            if (stripos($path, '?')) {
                $qloc = stripos($path, '?');
                $this->query = trim(substr($path, $qloc), '?');
                $path = substr($path, 0, $qloc);
            }
            $this->path = $path;
        }
        // get the body
        if ($getbody) {
            if (function_exists('http_get_request_body')) {
                $this->body = http_get_request_body();
            } else if (defined('STDIN')) {
                $this->body = @stream_get_contents(STDIN);
            } else {
                // Moodlerooms does not allow use of php://input
                // $this->body = @file_get_contents('php://input');
                // cannot get the body
                $this->set_header('NO_BODY', 'Cannot retrieve request body content');
                $this->body = null;
            }
        }
        // allow for method overrides
        $currentmethod = $_SERVER['REQUEST_METHOD'];
        if (isset($_REQUEST[self::COMPENSATE_METHOD])) {
            $compmethod = $_REQUEST[self::COMPENSATE_METHOD];
            if (! empty($compmethod)) {
                // Allows override to GET or DELETE
                $compmethod = strtoupper(trim($compmethod));
                if ('GET' == $compmethod) {
                    $currentmethod = 'GET';
                } else {
                    if ('DELETE' == $compmethod) {
                        $currentmethod = 'DELETE';
                    } else {
                        if ('POST' == $compmethod) {
                            $currentmethod = 'POST';
                        }
                    }
                }
            }
        }
        $this->method = $currentmethod;
    }

    public function set_status($status) {
        if ($status) {
            $this->status = $status;
        }
    }

    public function set_message($msg) {
        $this->message = $msg;
    }

    public function set_content_type($mimetype) {
        $this->headers['Content-Type'] = $mimetype;
    }

    public function set_header($name, $value) {
        $this->headers[$name] = $value;
    }

    /**
     * Send the response
     *
     * @param string $content [optional] the content to send
     * @param string $message [optional] the message to send, defaults to "Invalid request"
     */
    public function send_response($content = null, $message = null) {
        $code = $this->status;
        if (!isset($message)) {
            switch ($code) {
                case 200:
                    $message = 'OK';
                    break;
                case 204:
                    $message = 'No Content';
                    break;
                case 302:
                    $message = 'Found';
                    break;
                case 400:
                    $message = 'Bad Request';

                    break;
                case 401:
                    $message = 'Unauthorized';
                    break;
                case 403:
                    $message = 'Forbidden';
                    break;
                case 404:
                    $message = 'Not Found';
                    break;
                case 426:
                    $message = 'Upgrade Required';
                    break;
                default:
                    $message = 'Internal Server Error';
            }
        }
        header("HTTP/1.0 $code ".str_replace("\n", "", $message));
        if ($code >= 400) {
            // force plain text encoding when errors occur
            $this->set_content_type('text/plain');
        }
        $headers = $this->headers;
        if (isset($headers) && ! empty($headers)) {
            foreach ($headers as $key => &$value) {
                header($key.': '.$value, false);
            }
            unset($value);
        }
        // dump the body content
        if (isset($content)) {
            echo $content;
        }
    }

    // XHTML view processors
    public function process_registration() {
        // process calls to the registration view
        if (!iclicker_service::get_current_user_id()) {
            throw new ClickerSecurityException('Current user is not logged in and cannot access the registration view');
        }
        $this->results['new_reg'] = false;
        $this->results['clickeridval'] = "";
        if ('POST' == $this->method) {
            if (optional_param('register', null, PARAM_ALPHANUM) != null) {
                // we are registering a clicker
                $clickerid = optional_param('clickerId', null, PARAM_ALPHANUMEXT);
                if ($clickerid == null) {
                    $this->add_message(self::KEY_ERROR, 'reg.registered.clickerId.empty');
                } else {
                    $this->results['clickeridval'] = $clickerid;
                    // save a new clicker registration
                    try {
                        iclicker_service::create_clicker_registration($clickerid);
                        $this->add_message(self::KEY_INFO, 'reg.registered.success', $clickerid);
                        $this->add_message(self::KEY_BELOW, 'reg.registered.below.success');
                        $this->results['new_reg'] = true;
                    } catch (ClickerRegisteredException $e) {
                        $this->add_message(self::KEY_ERROR, 'reg.registered.clickerId.duplicate', $clickerid);
                        if (iclicker_service::$allowremotesharing) {
                            $this->add_message(self::KEY_BELOW, 'reg.registered.below.duplicate', $clickerid);
                        } else {
                            $this->add_message(self::KEY_BELOW, 'reg.registered.below.duplicate.noshare', $clickerid);
                        }
                    } catch (ClickerWebservicesException $e) {
                        $this->add_message(self::KEY_ERROR, 'reg.registered.clickerId.failure', $clickerid, $e);
                    } catch (ClickerIdInvalidException $e) {
                        if (ClickerIdInvalidException::F_EMPTY == $e->type) {
                            $this->add_message(self::KEY_ERROR, 'reg.registered.clickerId.empty');
                        } else if (ClickerIdInvalidException::F_LENGTH == $e->type) {
                            $this->add_message(self::KEY_ERROR, 'reg.registered.clickerId.wrong.length');
                        } else if (ClickerIdInvalidException::GO_NO_USER == $e->type) {
                            $this->add_message(self::KEY_ERROR, 'reg.registered.clickerId.failure', $clickerid, $e);
                        } else if (ClickerIdInvalidException::GO_LASTNAME == $e->type) {
                            $this->add_message(self::KEY_ERROR, 'reg.registered.clickerId.go.wrong.lastname', $e);
                        } else if (ClickerIdInvalidException::GO_NO_MATCH == $e->type) {
                            $this->add_message(self::KEY_ERROR, 'reg.registered.clickerId.go.invalid', $clickerid);
                        } else {
                            $this->add_message(self::KEY_ERROR, 'reg.registered.clickerId.invalid', $clickerid);
                        }
                    }
                }

            } else {
                if (optional_param('activate', null, PARAM_ALPHANUM) != null) {
                    // First arrived at this page
                    $activate = optional_param('activate', 'false', PARAM_ALPHANUM);
                    $activate = ($activate == 'true' ? true : false);
                    $regid = optional_param('registrationId', null, PARAM_INT);
                    if ($regid == null) {
                        $this->add_message(self::KEY_ERROR, 'reg.activate.registrationId.empty', null);
                    } else {
                        // save a new clicker registration
                        $cr = iclicker_service::set_registration_active($regid, $activate);
                        if ($cr) {
                            $this->add_message(self::KEY_INFO, 'reg.activate.success.' . ($cr->activated ? 'true' : 'false'), $cr->clickerid);
                        }
                    }

                } else {
                    if (optional_param('remove', null, PARAM_ALPHANUM) != null) {
                        $regid = optional_param('registrationId', null, PARAM_ALPHANUMEXT);
                        if (($regid == null)) {
                            $this->add_message(self::KEY_ERROR, 'reg.activate.registrationId.empty', null);
                        } else {
                            // remove a new clicker registration by deactivating it
                            $cr = iclicker_service::set_registration_active($regid, false);
                            if ($cr) {
                                $this->add_message(self::KEY_INFO, 'reg.remove.success', $cr->clickerid);
                            }
                        }

                    } else {
                        // invalid POST
                        echo('WARN: Invalid POST: does not contain register or activate, nothing to do');
                    }
                }
            }
        }

        $this->results['regs'] = iclicker_service::get_registrations_by_user(null, true);
        $this->results['is_instructor'] = iclicker_service::is_instructor();
        $this->results['ssoenabled'] = iclicker_service::$blockiclickerssoenabled;
        // added to allow special messages below the forms
        $this->results['belowmessages'] = $this->get_messages(self::KEY_BELOW);
    }

    public function process_instructor() {
        $this->results['instpath'] = iclicker_service::block_url('instructor.php');
        // admin/instructor check
        if (!iclicker_service::is_admin() && !iclicker_service::is_instructor()) {
            throw new ClickerSecurityException("Current user is not an instructor and cannot access the instructor view");
        }
        $courseid = optional_param('courseId', false, PARAM_INT);
        $this->results['course_id'] = $courseid;
        $courses = array();
        if ($courseid) {
            $course = iclicker_service::get_course($courseid);
            $courses[] = $course;
        } else {
            $courses = iclicker_service::get_courses_for_instructor();
        }
        $this->results['courses'] = $courses;
        $this->results['coursescount'] = count($courses);
        $this->results['showstudents'] = false;
        if ($courseid && count($courses) == 1) {
            $course = $courses[0];
            $this->results['showstudents'] = true;
            $this->results['course'] = $course;
            $students = iclicker_service::get_students_for_course_with_regs($courseid);
            $this->results['students'] = $students;
            $this->results['studentscount'] = count($students);
        }
        $this->results['ssoenabled'] = iclicker_service::$blockiclickerssoenabled;
    }

    public function process_instructor_sso() {
        $this->results['instpath'] = iclicker_service::block_url('instructor.php');
        // admin/instructor check
        if (!iclicker_service::is_admin() && !iclicker_service::is_instructor()) {
            throw new ClickerSecurityException("Current user is not an instructor and cannot access the instructor view");
        }
        $this->results['ssoenabled'] = iclicker_service::$blockiclickerssoenabled;
        $currentuserid = iclicker_service::get_current_user_id();
        if (iclicker_service::$blockiclickerssoenabled) {
            $currentuserkey = null;
            if ('POST' == $this->method) {
                if ( optional_param('generateKey', false, PARAM_ALPHANUM) != null ) {
                    // handle generating a new key
                    $currentuserkey = iclicker_service::make_user_key($currentuserid, true);
                    $this->add_message(self::KEY_INFO, 'inst.sso.generated.new.key', null);
                }
            }
            if ($currentuserkey == null) {
                $currentuserkey = iclicker_service::make_user_key($currentuserid, false);
            }
            $this->results['ssouserkey'] = $currentuserkey;
        }
        // $currentuserkey = iclicker_service::make_user_key($currentuserid);
        // $this->results['ssouserkey'] = $currentuserkey;
    }

    public function process_admin() {
        global $CFG;
        $adminpath = iclicker_service::block_url('admin.php');
        $this->results['adminpath'] = $adminpath;
        $this->results['status_url'] = iclicker_service::block_url('runner_status.php');
        // admin check
        if (!iclicker_service::is_admin()) {
            throw new ClickerSecurityException("Current user is not an admin and cannot access the admin view");
        }

        // get sorting params
        $pagenum = 1;
        $perpagenum = 20; // does not change
        if (optional_param('page', null, PARAM_ALPHANUM) != null) {
            $pagenum = required_param('page', PARAM_INT);
            if ($pagenum < 1) {
                $pagenum = 1;
            }
        }

        $sort = 'clickerid';
        if (optional_param('sort', null, PARAM_ALPHANUM) != null) {
            $sort = required_param('sort', PARAM_ALPHANUMEXT);
        }
        $this->results['sort'] = $sort;

        // get filtering params
        $search = optional_param('search', null, PARAM_ALPHANUMEXT);
        if (empty($search)) {
            $search = null;
        }

        $startdate = null;
        $startdatetext = '';
        if (optional_param('start_date', null, PARAM_ALPHANUMEXT) != null) {
            $startdate = required_param('start_date', PARAM_ALPHANUMEXT);
            $startdate = strtotime($startdate);
            if ($startdate) {
                $startdatetext = date('Y-m-d', $startdate);
            } else {
                $startdate = null;
            }
        }

        $enddate = null;
        $enddatetext = '';
        if (optional_param('end_date', null, PARAM_ALPHANUMEXT) != null) {
            $enddate = required_param('end_date', PARAM_ALPHANUMEXT);
            $enddate = strtotime($enddate);
            if ($enddate) {
                // need to make this the end of the day
                $enddt = new DateTime('@' . $enddate);
                $enddt->setTimezone(new DateTimeZone(date_default_timezone_get()));
                $enddt->setTime(23, 59, 59);
                $enddate = $enddt->getTimestamp();
                $enddatetext = date('Y-m-d', $enddate);
            } else {
                $enddate = null;
            }
        }

        if ("POST" == $this->method) {
            if (optional_param('activate', null, PARAM_ALPHANUM) != null) {
                // First arrived at this page
                $activate = required_param('activate', PARAM_BOOL);
                if (optional_param('registrationId', null, PARAM_ALPHANUMEXT) == null) {
                    $this->add_message(self::KEY_ERROR, "reg.activate.registrationId.empty", null);
                } else {
                    $regid = required_param('registrationId', PARAM_INT);
                    // save a new clicker registration
                    $cr = iclicker_service::set_registration_active($regid, $activate);
                    if ($cr) {
                        $args = new stdClass;
                        $args->cid = $cr->clickerid;
                        $args->user = iclicker_service::get_user_displayname($cr->ownerid);
                        $this->add_message(self::KEY_INFO, "admin.activate.success.".($cr->activated ? 'true' : 'false'), $args);
                    }
                }

            } else if (optional_param('remove', null, PARAM_ALPHANUM) != null) {
                if (optional_param('registrationId', null, PARAM_ALPHANUMEXT) == null) {
                    $this->add_message(self::KEY_ERROR, "reg.activate.registrationId.empty", null);
                } else {
                    $regid = required_param('registrationId', PARAM_INT);
                    $cr = iclicker_service::get_registration_by_id($regid);
                    if ($cr) {
                        iclicker_service::remove_registration($regid);
                        $args = new stdClass();
                        $args->cid = $cr->clickerid;
                        $args->rid = $regid;
                        $args->user = iclicker_service::get_user_displayname($cr->ownerid);
                        $this->add_message(self::KEY_INFO, "admin.delete.success", $args);
                    }
                }

            } else if (optional_param('purge', null, PARAM_ALPHANUM) != null) {
                // actually do the purging
                $count = iclicker_service::purge_registrations($search, $startdate, $enddate);
                $this->add_message(self::KEY_INFO, "admin.purge.success", $count);
                // reset all search params to defaults
                $search = null;
                $startdate = null;
                $enddate = null;
                $pagenum = 1;
                $perpagenum = 20;
                $sort = 'clickerid';

            } else {
                // invalid POST
                debugging('WARN: Invalid POST: does not contain remove, purge, or activate, nothing to do');
            }
        }

        // put search and sort data into the page
        $this->results['search'] = $search;
        $this->results['startdate'] = $startdate;
        $this->results['startdateText'] = $startdatetext;
        $this->results['endDate'] = $enddate;
        $this->results['endDateText'] = $enddatetext;
        $this->results['page'] = $pagenum;
        $this->results['perPage'] = $perpagenum;
        $this->results['sort'] = $sort;

        // put config data into page
        $this->results['ssoenabled'] = iclicker_service::$blockiclickerssoenabled;
        $this->results['ssosharedkey'] = iclicker_service::$blockiclickerssosharedkey;
        $this->results['domainURL'] = iclicker_service::$domainurl;
        $this->results['adminemailaddress'] = empty($CFG->block_iclicker_notify_emails) ? '' : $CFG->block_iclicker_notify_emails;

        // put error data into page
        $this->results['recentfailures'] = iclicker_service::get_failures();

        // handling the calcs for paging
        $first = ($pagenum - 1) * $perpagenum;
        $totalcount = iclicker_service::count_all_registrations($search, $startdate, $enddate);
        $pagecount = floor(($totalcount + $perpagenum - 1) / $perpagenum);
        $this->results['totalcount'] = $totalcount;
        $this->results['page_count'] = $pagecount;
        $this->results['registrations'] = iclicker_service::get_all_registrations($first, $perpagenum, $sort, $search, $startdate, $enddate);

        $pagerhtml = "";
        if ($totalcount > 0) {
            $timestamp = microtime();
            for ($i = 0; $i < $pagecount; $i++) {
                $currentpage = $i + 1;
                $currentstart = ($i * $perpagenum) + 1;
                $currentend = $currentstart + $perpagenum - 1;
                if ($currentend > $totalcount) {
                    $currentend = $totalcount;
                }
                $marker = '['.$currentstart.'..'.$currentend.']';
                if ($currentpage == $pagenum) {
                    // make it bold and not a link
                    $pagerhtml .= '<span class="paging_current paging_item">'.$marker.'</span>'."\n";
                } else {
                    // make it a link
                    $pagingurl = $adminpath.'?page='.$currentpage.'&sort='.$sort;
                    if (isset($search)) {
                        $pagingurl .= '&search='.urlencode($search);
                    }
                    if (!empty($startdatetext)) {
                        $pagingurl .= '&start_date='.urlencode($startdatetext);
                    }
                    if (!empty($enddatetext)) {
                        $pagingurl .= '&end_date='.urlencode($enddatetext);
                    }
                    $pagingurl .= '&nc='.($timestamp.$currentpage);
                    $pagerhtml .= '<a class="paging_link paging_item" href="'.$pagingurl.'">'.$marker.'</a>'."\n";
                }
            }
            $this->results['pagerhtml'] = $pagerhtml;
        }
    }

    public function process_admin_csv() {
        // global $CFG;
        // admin check
        if (!iclicker_service::is_admin()) {
            throw new ClickerSecurityException("Current user is not an admin and cannot make CSV of all regs");
        }
        $this->results['registrations'] = iclicker_service::get_all_registrations();
    }

    // MESSAGING

    public $messages = array();

    const KEY_INFO = "INFO";
    const KEY_ERROR = "ERROR";
    const KEY_BELOW = "BELOW";

    /**
     * Adds a message
     *
     * @param string $key the KEY const
     * @param string $message the message to add
     * @param string $extra [OPTIONAL] extra string to include hidden on the page with the message
     * @throws Exception if the key is invalid
     */
    public function add_message_str($key, $message, $extra = null) {
        if ($key == null) {
            throw new Exception("key (".$key.") must not be null");
        }
        if ($message) {
            if (!isset($this->messages[$key])) {
                $this->messages[$key] = array(
                );
            }
            if (isset($extra)) {
                if (!is_string($extra)) {
                    $extra = var_export($extra, true);
                }
                $extra = PHP_EOL . '<div style="display:none;">' . $extra . '</div>' . PHP_EOL;
            } else {
                $extra = '';
            }
            $this->messages[$key][] = $message . $extra;
        }
    }

    /**
     * Add an i18n message based on a key
     *
     * @param string $key the KEY const
     * @param string $messagekey the i18n message key
     * @param object $args [OPTIONAL] args to include
     * @param string $extra [OPTIONAL] extra string to include hidden on the page with the message
     * @throws Exception if the key is invalid
     */
    public function add_message($key, $messagekey, $args = null, $extra = null) {
        if ($key == null) {
            throw new Exception("key (".$key.") must not be null");
        }
        if ($messagekey) {
            $message = iclicker_service::msg($messagekey, $args);
            $this->add_message_str($key, $message, $extra);
        }
    }

    /**
     * Get the messages that are currently waiting in this request
     *
     * @param string $key the KEY const
     * @return array the list of messages to display
     * @throws Exception if the key is invalid
     */
    public function get_messages($key) {
        if ($key == null) {
            throw new Exception("key (".$key.") must not be null");
        }
        $messages = null;
        if (isset($this->messages[$key])) {
            $messages = $this->messages[$key];
            if (!isset($messages)) {
                $messages = array(
                );
            }
        } else {
            $messages = array(
            );
        }
        return $messages;
    }

}
