<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Version details
 *
 * @package    block_iclicker
 * @copyright  2012 i>clicker (R) <http://www.iclicker.com/dnn/> onwards Saji Nair (http://sajinair.com)
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

// This includes lib/setup.php and the standard set:
// setup.php : setup which creates the globals
// '/textlib.class.php');   // Functions to handle multibyte strings
// '/weblib.php');          // Functions for producing HTML
// '/dmllib.php');          // Functions to handle DB data (DML) - inserting, updating, and retrieving data from the database
// '/datalib.php');         // Legacy lib with a big-mix of functions. - user, course, etc. data lookup functions
// '/accesslib.php');       // Access control functions - context, roles, and permission related functions
// '/deprecatedlib.php');   // Deprecated functions included for backward compatibility
// '/moodlelib.php');       // general-purpose (login, getparams, getconfig, cache, data/time)
// '/eventslib.php');       // Events functions
// '/grouplib.php');        // Groups functions.

// Ddlib.php : modifying, creating, or deleting database schema
// Blocklib.php : functions to use blocks in a typical course page
// Formslib.php : classes for creating forms in Moodle, based on PEAR QuickForms.

require_once('../../config.php');
global $CFG, $USER, $COURSE;
require_once('iclicker_service.php');
require_once('controller.php');


// INTERNAL METHODS.
/**
 * This will check for a user and return the userid if one can be found
 * @param string $msg the error message
 * @return int the userid
 * @throws ClickerSecurityException if no user can be found
 */
function iclicker_get_and_check_current_user($msg) {
    $userid = iclicker_service::get_current_user_id();
    if (! $userid) {
        throw new ClickerSecurityException("Only logged in users can $msg");
    }
    if (! iclicker_service::is_admin($userid) && ! iclicker_service::is_instructor($userid)) {
        throw new ClickerSecurityException("Only instructors can " . $msg);
    }
    return $userid;
}

/**
 * Attempt to authenticate the current request based on request params and basic auth
 * @param iclicker_controller $cntlr the controller instance
 * @throws ClickerSecurityException if authentication is impossible given the request values
 * @throws ClickerSSLRequiredException if the auth request is bad (requires SSL but SSL not used)
 */
function iclicker_handle_authn($cntlr) {
    global $CFG;
    // extract the authn params
    $authusername = optional_param(iclicker_controller::LOGIN, null, PARAM_NOTAGS);
    $authpassword = optional_param(iclicker_controller::PASSWORD, null, PARAM_NOTAGS);
    if (empty($authusername) && isset($_SERVER['PHP_AUTH_USER'])) {
        // No username found in normal params so try to get basic auth.
        $authusername = $_SERVER['PHP_AUTH_USER'];
        $authpassword = $_SERVER['PHP_AUTH_PW'];
        if (empty($authusername)) {
            // Attempt to get it from the header as a final try.
            list($authusername, $authpassword) = explode(':', base64_decode(substr($_SERVER['HTTP_AUTHORIZATION'], 6)));
        }
    }
    if (iclicker_service::$blockiclickerssoenabled && !empty($authpassword)) {
        // When SSO is enabled and the password is set it means this is not actually a user password so we can proceed without requiring SSL.
        $null = null;
    } else {
        // This is a user password so https must be used if the loginhttps option is enabled.
        $sslrequest = (!empty($_SERVER['HTTPS']) && $_SERVER['HTTPS'] !== 'off') || $_SERVER['SERVER_PORT'] == 443;
        $sslrequired = (isset($CFG->forcehttps) && $CFG->forcehttps == true) || (isset($CFG->loginhttps) && $CFG->loginhttps == true);
        if ($sslrequired && !$sslrequest) {
            throw new ClickerSSLRequiredException('SSL is required when performing a user login (and sending user passwords)');
        }
    }
    // $session_id = optional_param(iclicker_controller::SESSION_ID, NULL, PARAM_NOTAGS);
    if (!empty($authusername)) {
        $ssokey = optional_param(iclicker_controller::SSO_KEY, null, PARAM_NOTAGS);
        iclicker_service::authenticate_user($authusername, $authpassword, $ssokey); // throws exception if fails
        // } Else if ($session_id) {
        // $valid = FALSE; // validate the session key
        // if (! $valid) {
        // throw new SecurityException("Invalid "+iclicker_controller::SESSION_ID+" provided, session may have expired, send new login credentials");
        // }
    }
    $currentuserid = iclicker_service::get_current_user_id();
    if (isset($currentuserid)) {
        $cntlr->set_header(iclicker_controller::SESSION_ID, sesskey());
        $cntlr->set_header('_userId', $currentuserid);
    }
}

/**
 * Extracts the XML data from the request
 * @param object $cntlr the controller instance
 * @return string the XML data OR null if none can be found
 */
function iclicker_get_xml_data($cntlr) {
    $xml = optional_param(iclicker_controller::XML_DATA, null, PARAM_RAW_TRIMMED);
    if (empty($xml)) {
        $xml = $cntlr->body;
    } else {
        $xml = stripslashes($xml);
    }
    return $xml;
}


// REST HANDLING
// require_login();
// echo "me=".me().", qualified=".qualified_me();
// echo "user: id=".$USER->id.", auth=".$USER->auth.", username=".$USER->username.", lastlogin=".$USER->lastlogin."\n";
// echo "course: id=".$COURSE->id.", title=".$COURSE->fullname."\n";
// echo "CFG: wwwroot=".$CFG->wwwroot.", httpswwwroot=".$CFG->httpswwwroot.", dirroot=".$CFG->dirroot.", libdir=".$CFG->libdir."\n";.

// Activate the controller.
$cntlr = new iclicker_controller(true); // with body

// init the vars to success
$valid = true;
$status = 200; // ok
$output = '';

// check to see if this is one of the paths we understand
if (! $cntlr->path) {
    $valid = false;
    $output = "Unknown path ($cntlr->path) specified";
    $status = 404; // not found
}
if ($valid
        && "POST" != $cntlr->method
        && "GET" != $cntlr->method) {
    $valid = false;
    $output = "Only POST and GET methods are supported";
    $status = 405; // method not allowed
}
if ($valid) {
    // check against the ones we know and process
    $parts = explode('/', $cntlr->path);
    $pathseg0 = count($parts) > 0 ? $parts[0] : null;
    $pathseg1 = count($parts) > 1 ? $parts[1] : null;
    $pathseg2 = count($parts) > 2 ? $parts[2] : null;
    try {
        if ($pathseg0 == 'verifykey') {
            // SPECIAL case handling (no authn handling)
            $ssokey = optional_param(iclicker_controller::SSO_KEY, null, PARAM_NOTAGS);
            if (iclicker_service::verify_key($ssokey)) {
                $cntlr->set_status(200);
                $output = "Verified";
            } else {
                $cntlr->set_status(501);
                $output = "Disabled";
            }
            $cntlr->set_content_type("text/plain");
            $cntlr->send_response($output);
            return;

            /*
            } else if ('verify_go_ws' == $pathseg0) {
                // test the go webservices
                $clickerid = $pathseg1;
                $last_name = $pathseg2;
                $output = iclicker_service::ws_go_verify_clickerid($clickerid, $last_name);
                $cntlr->set_content_type("text/plain");
                $cntlr->send_response($output);
                return;
                */

        } else {
            // NORMAL case handling
            // handle the request authn if needed
            iclicker_handle_authn($cntlr);
            if ("GET" == $cntlr->method) {
                if ("courses" == $pathseg0) {
                    // handle retrieving the list of courses for an instructor
                    $userid = iclicker_get_and_check_current_user("access instructor courses listings");
                    $output = iclicker_service::encode_courses($userid);

                } else if ("students" == $pathseg0) {
                    // handle retrieval of the list of students
                    $courseid = $pathseg1;
                    if ($courseid == null) {
                        throw new InvalidArgumentException(
                                "valid course_id must be included in the URL /students/{course_id}");
                    }
                    iclicker_get_and_check_current_user("access student enrollment listings");
                    $output = iclicker_service::encode_enrollments($courseid);

                } else {
                    // UNKNOWN
                    $valid = false;
                    $output = "Unknown path ($cntlr->path) specified for method GET";
                    $status = 404; // NOT_FOUND
                }
            } else {
                // POST
                if ("gradebook" == $pathseg0) {
                    // Handle retrieval of the list of students.
                    $courseid = $pathseg1;
                    if ($courseid == null) {
                        throw new InvalidArgumentException(
                                "valid course_id must be included in the URL /gradebook/{course_id}");
                    }
                    iclicker_get_and_check_current_user("upload grades into the gradebook");
                    $xml = iclicker_get_xml_data($cntlr);
                    try {
                        $gradebook = iclicker_service::decode_gradebook($xml);
                        // Process gradebook data.
                        $results = iclicker_service::save_gradebook($gradebook);
                        // Generate the output.
                        $output = iclicker_service::encode_gradebook_results($results);
                        if (! $output) {
                            // Special RETURN, non-XML, no failures in save.
                            $cntlr->set_status(200);
                            $cntlr->set_content_type("text/plain");
                            $output = "True";
                            $cntlr->send_response($output);
                            return; // SHORT CIRCUIT.
                        } else {
                            // Failures occurred during save.
                            $status = 200; // OK;
                        }
                    } catch (InvalidArgumentException $e) {
                        // Invalid XML.
                        $valid = false;
                        $output = "Invalid gradebook XML in request, unable to process:/n $xml";
                        $status = 400; // BAD_REQUEST;
                    }

                } else if ("authenticate" == $pathseg0) {
                    iclicker_get_and_check_current_user("authenticate via iclicker");
                    // Special return, non-XML.
                    $cntlr->set_status(204); // No content.
                    $cntlr->send_response();
                    return; // SHORT CIRCUIT.

                } else if ("register" == $pathseg0) {
                    iclicker_get_and_check_current_user("upload registrations data");
                    $xml = iclicker_get_xml_data($cntlr);
                    $cr = iclicker_service::decode_registration($xml);
                    $ownerid = $cr->ownerid;
                    $message = '';
                    $regstatus = false;
                    try {
                        iclicker_service::create_clicker_registration($cr->clickerid, $ownerid);
                        // Valid registration.
                        $message = iclicker_service::msg('reg.registered.below.success', $cr->clickerid);
                        $regstatus = true;
                    } catch (ClickerIdInvalidException $e) {
                        // Invalid clicker id.
                        $message = iclicker_service::msg('reg.registered.clickerId.invalid', $cr->clickerid);
                    } catch (InvalidArgumentException $e) {
                        // Invalid user id.
                        $message = "Student not found in the CMS";
                    } catch (ClickerRegisteredException $e) {
                        // already registered
                        $key = '';
                        if ($e->ownerid == $e->registeredownerid) {
                            // already registered to this user
                            $key = 'reg.registered.below.duplicate';
                        } else {
                            // already registered to another user
                            $key = 'reg.registered.clickerId.duplicate.notowned';
                        }
                        $message = iclicker_service::msg($key, $cr->clickerid);
                    }
                    $registrations = iclicker_service::get_registrations_by_user($ownerid, true);
                    $output = iclicker_service::encode_registration_result($registrations, $regstatus, $message);
                    if ($regstatus) {
                        $status = 200; // OK;
                    } else {
                        $status = 400; // BAD_REQUEST;
                    }

                } else {
                    // UNKNOWN.
                    $valid = false;
                    $output = "Unknown path ($cntlr->path) specified for method POST";
                    $status = 404; // NOT_FOUND;
                }
            }
        }
    } catch (ClickerSecurityException $e) {
        $valid = false;
        $currentuserid = iclicker_service::get_current_user_id();
        if (! $currentuserid) {
            $output = "User must be logged in to perform this action: " . $e;
            $status = 403; // UNAUTHORIZED;
        } else {
            $output = "User ($currentuserid) is not allowed to perform this action: " . $e;
            $status = 401; // FORBIDDEN;
        }
    } catch (InvalidArgumentException $e) {
        $valid = false;
        $output = "Invalid request: " . $e;
        // Log.warn("i>clicker: " + $output, $e);
        $status = 400; // BAD_REQUEST;
    } catch (ClickerSSLRequiredException $e) {
        $valid = false;
        $output = "SSL_REQUIRED: " . $e;
        // Log.warn("i>clicker: " + $output, $e);
        $status = 426; // UPGRADE_REQUIRED;
    } catch (Exception $e) {
        $valid = false;
        $output = "Failure occurred: " . $e;
        // Log.warn("i>clicker: " + $output, $e);
        $status = 500; // INTERNAL_SERVER_ERROR;
    }
}
if ($valid) {
    // Send the response.
    $cntlr->set_status(200);
    $cntlr->set_content_type('application/xml');
    $output = iclicker_controller::XML_HEADER . $output;
    $cntlr->send_response($output);
} else {
    // error with info about how to do it right
    $cntlr->set_status($status);
    $cntlr->set_content_type('text/plain');
    // add helpful info to the output
    $msg = "ERROR $status: Invalid request (".$cntlr->method." /".$cntlr->path.")" .
        "\n\n=INFO========================================================================================\n".
        $output.
        "\n\n-HELP----------------------------------------------------------------------------------------\n".
        "Valid request paths include the following (without the block prefix: ".iclicker_service::block_url('rest.php')."):\n".
        "POST /authenticate             - authenticate by sending credentials (".iclicker_controller::LOGIN.",".iclicker_controller::PASSWORD.") \n".
        "                                 return status 204 (valid login) \n".
        "POST /verifykey                - check the encoded key is valid and matches the shared key \n".
        "                                 return 200 if valid OR 501 if SSO not enabled OR 400/401 if key is bad \n".
        "POST /register                 - Add a new clicker registration, return 200 for success or 400 with \n".
        "                                 registration response (XML) for failure \n".
        "GET  /courses                  - returns the list of courses for the current user (XML) \n".
        "GET  /students/{course_id}     - returns the list of student enrollments for the given course (XML) \n".
        "                                 or 403 if user is not an instructor in the specified course \n".
        "POST /gradebook/{course_id}    - send the gradebook data into the system, returns errors on failure (XML) \n".
        "                                 or 'True' if no errors, 400 if the xml is missing or course_id is invalid, \n".
        "                                 403 if user is not an instructor in the specified course \n".
        "\n".
        " - Authenticate by sending credentials (".iclicker_controller::LOGIN.",".iclicker_controller::PASSWORD.") in the request parameters \n".
        " -- SSO authentication requires an encoded key (".iclicker_controller::SSO_KEY.") in the request parameters \n".
        " -- The response headers will include the sessionId when credentials are valid \n".
        " -- Invalid credentials or sessionId will result in a 401 (invalid credentials) or 403 (not authorized) status \n".
        " - Use ".iclicker_controller::COMPENSATE_METHOD." param to override the http method being used (e.g. POST /courses?".iclicker_controller::COMPENSATE_METHOD."=GET will".
        "   force the method to be a GET despite sending as a POST) \n".
        " - Send data as the http request BODY or as a form parameter called ".iclicker_controller::XML_DATA." \n".
        " - All endpoints return 403 if user is not an instructor \n".
        " - Version: ".iclicker_service::VERSION." (".iclicker_service::BLOCK_VERSION.") \n";
    $cntlr->send_response($msg);
}
